/************************************************************************
* Copyright (c) 1995 Valentin Shopov, Blagoevgrad, Bulgaria             *
*                                                                       *
* Permission to use, copy, modify, and distribute this material for any *
* purpose and without fee is hereby granted, provided that the above    *
* copyright notice and this permission notice appear in all copies, and *
* that the name of the author not be used in advertising or publicity   *
* pertaining to this material without the specific, prior written       *
* permission of the author.                                             *
*                                                                       *
* THE AUTHOR MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY *
* OF THIS MATERIAL FOR ANY PURPOSE. IT IS PROVIDED "AS IS", WITHOUT ANY *
* EXPRESS OR IMPLIED WARRANTIES.                                        *
*************************************************************************/

/**************************ciilib1.5.1***********************************
 *                                                                      *
 *        Character to Inline Image Conversion Library (CIILIB)         *
 *         This version of CIILIB with the following functions 		*
 *			by Mr. Akira Kitauchi				*
 *	improves availability and speed:				*
 *       - serves CII with the image/GIF content type			*
 *       - serves CII word by word (not character by character)		*
 *									*
 *                                                                      *
 *    GIF coded CII							*
 *              August 1995, by Akira KITAUCHI				*
 *									*
 *              E-mail address: kitauchi@i.h.kyoto-u.ac.jp		*
 *              WWW home page : http://www.i.h.kyoto-u.ac.jp/~kitauchi/	*
 ************************************************************************/
/*
   #include <memory.h>
 */
#define fgColor	(isAnchor ? lkColor : txColor)
#define GIFW ((chW>8)? ((GIFWIDTH<2)? 2 : (GIFWIDTH/2)*2) : ((GIFWIDTH<2)?1:GIFWIDTH))
#define GIFWH1 ((chW>8)? ((GIFWIDTH<4)? 2 : (GIFWIDTH/4)*2) : ((GIFWIDTH<2)?1:GIFWIDTH/2))

static int isH2, isH3, isH4, isH5, isH6, isXBM;
static int txColor = 0x000000;
static int lkColor = 0x0000ee;
static char *xbmname;

/*----
 *	text to CII/GIF converter
 */
static void
GIFCII_init (isMozilla)
     int isMozilla;
{
  if (!*Hchar)
    Hchar = "k";
  if (!isMozilla)
    lkColor = 0x0000b0;
}

static
GIFCII_setfsize (chHp, chWp, Acharp, charset)
     int *chHp, *chWp;
     char **Acharp;
     char *charset;
{
  *chHp = CII_CC[CharsetID (charset)].height;
  *chWp = CII_CC[CharsetID (charset)].width;

#ifdef UNDERLINE
  if (isAnchor && isMozilla)
    {
      *Acharp = "u";
      *chHp += 2;
    }
#endif
}

static int
GIFCII_switch_H (sp)
     char *sp;
{
  int match = 1;

  if (!isH2 && !strncscmp (sp, "<H2>", 4))
    {
      isH2 = 1;
      Hchar = "i";
    }
  else if (!isH3 && !strncscmp (sp, "<H3>", 4))
    {
      isH3 = 1;
      Hchar = "j";
    }
  else if (!isH4 && !strncscmp (sp, "<H4>", 4))
    {
      isH4 = 1;
      Hchar = "j";
    }
  else if (!isH5 && !strncscmp (sp, "<H5>", 4))
    {
      isH5 = 1;
      Hchar = "j";
    }
  else if (!isH6 && !strncscmp (sp, "<H6>", 4))
    {
      isH6 = 1;
      Hchar = "j";
    }
  else if (isH1 && !strncscmp (sp, "</H1>", 5))
    {
      isH1 = 0;
      Hchar = "k";
    }
  else if (isH2 && !strncscmp (sp, "</H2>", 5))
    {
      isH2 = 0;
      Hchar = "k";
    }
  else if (isH3 && !strncscmp (sp, "</H3>", 5))
    {
      isH3 = 0;
      Hchar = "k";
    }
  else if (isH4 && !strncscmp (sp, "</H4>", 5))
    {
      isH4 = 0;
      Hchar = "k";
    }
  else if (isH5 && !strncscmp (sp, "</H5>", 5))
    {
      isH5 = 0;
      Hchar = "k";
    }
  else if (isH6 && !strncscmp (sp, "</H6>", 5))
    {
      isH6 = 0;
      Hchar = "k";
    }
  else
    match = 0;
  return match;
}
static int
getColor (colorp, sp)
     int *colorp;
     char *sp;
{
  int i = 6, j, color = 0;

  while (sp[i] == ' ' || sp[i] == '\t')
    {
      i++;
    }
  if (sp[i] == '\"')
    {
      i++;
    }
  if (sp[i] == '#')
    {
      i++;
    }

  for (j = i; j < i + 6; j++)
    {
      if (sp[j] == 0 || sp[j] == ' ' || sp[j] == '\"' || sp[j] == '\n')
	break;

      color *= 0x10;

      if (sp[j] >= '0' && sp[j] <= '9')
	color += sp[j] - '0';
      else if (sp[j] >= 'A' && sp[j] <= 'F')
	color += sp[j] - 'A' + 10;
      else if (sp[j] >= 'a' && sp[j] <= 'f')
	color += sp[j] - 'a' + 10;
    }

  *colorp = color;
  return j - 1;
}
static int
GIFCII_switch_Color (sp, isMozilla)
     char *sp;
     int isMozilla;
{
  int i;
  if (isMozilla && !strncscmp (sp, "<BODY", 5))
    {
      int isText = 0, isLink = 0, j;
      for (i = 0; sp[i] != '>' && sp[i]; i++)
	{
	  if (sp[i] == ' ' || sp[i] == '\"' || sp[i] == '\n')
	    {
	      if (!isText && !strncscmp (sp + i + 1, "TEXT=", 5))
		{
		  isText = 1;
		  i += getColor (&txColor, sp + i);
		}
	      if (!isLink && !strncscmp (sp + i + 1, "LINK=", 5))
		{
		  isLink = 1;
		  i += getColor (&lkColor, sp + i);
		}
	    }
	}
      return 1;
    }
  else
    return 0;
}

static void
shrink_PIC (expbuf, picbuf, chN, chW, chH)
     char *expbuf, *picbuf;
     int chN, chW, chH;
{
  int h, n, w, isBRK;
  memset (picbuf, 0, ((chN * chW + 7) / 8) * chH);
  for (h = 0; h < chH; h++)
    {
      for (n = 0, isBRK = 0;; n++)
	{
	  for (w = 0; w < 8; w++)
	    {
	      if (8 * n + w + 1 > chN * chW)
		{
		  isBRK = (w == 0) ? 2 : 1;
		  break;
		}
	      *picbuf |= (*expbuf << w);
	      expbuf++;
	    }
	  picbuf++;
	  if (isBRK)
	    {
	      picbuf -= (isBRK - 1);
	      break;
	    }
	}
    }
}



static unsigned char *
GIFCII_putCII (dp, sp, chH, chW, ubase, charset, Achar, isMozilla, do256)
     unsigned char *dp, *sp, *ubase, *charset, *Achar;
{
  unsigned char t256c;
  int w;



  sprintf (dp, "<IMG SRC=\"%s%s/%06x%s%s", ubase, charset, fgColor, Hchar, Achar);
  dp += strlen (dp);

  for (w = 0; w < ((isH1) ? GIFWH1 : GIFW) && (sp[0] & 0x80) && ((do256) ? 1 : (sp[1] & 0x80)); w++)
    {
      t256c = (do256 && isKOI == 1) ? koi8tocyr (sp[0]) : sp[0];
      sprintf (dp, "%02x%02x", ((do256) ? 0 : t256c), ((do256) ? t256c : sp[1]));
      sp += (do256) ? 1 : 2;
      dp += 4;
      w += (chW > 8) ? 1 : 0;

#ifdef NOLONG256C
      if (do256 && !isAnchor && *Hchar == 'k')
	{
	  w++;
	  break;
	}
#endif

#ifdef LONGKANJIONLY
      if (isJP && !isAnchor && *Hchar == 'k')
	{
	  if ((t256c >= 0xa1 && t256c <= 0xa8) || (sp[0] >= 0xa1 && sp[0] <= 0xa8 && sp[1] & 0x80))
	    {
	      w++;
	      break;
	    }
	}
#endif

#ifdef USPACE
      if (isAnchor && *sp == ' ')
	{
	  for (; (w + 1) < ((isH1) ? GIFWH1 : GIFW) && *sp == ' '; w++)
	    {
	      sprintf (dp, "USPC");
	      sp++;
	      dp += 4;
	    }
	}
#endif

    }
  if (isMozilla)
    {
      if (chW > 8)
	chW /= 2;
      if (isH1)
	{
	  chW <<= 1;
	  chH <<= 1;
	}
      sprintf (dp, "\" WIDTH=%d HEIGHT=%d BORDER=0>", chW * w, chH);
    }
  else
    sprintf (dp, "\">");
  return sp;
}

/*----
 *	CII/GIF font server
 */
static void
expand_H1 (srcb, dstb, width, chW)
     char *srcb, *dstb;
     int width;
{
  int expb, expb0, dmask, smask, i, j;

  expb0 = 0;

  for (i = 0; i < width; i++)
    {
      expb = expb0;
      dmask = 7;
      smask = 1;
      for (j = 0; j < chW; j++)
	{
	  expb |= (srcb[i] & smask) ? dmask : 0;
	  smask <<= 1;
	  dmask <<= 2;
	}
      dstb[i * 2 + 0] = (char) expb;
      dstb[i * 2 + 1] = (char) (expb >> 8) & (0xff >> (16 - 2 * chW));
      expb0 = expb >> chW * 2;
    }
}


static int
PIC_to_XBM (picbuf, xbmbuf, chN, chW, chH, color)
     char *picbuf, *xbmbuf;
     int chN, chW, chH;
     int color;
{
  int i;
  char *xbm;
  char expbuf[32 * 34 * (GIFWIDTH + 1)];
  if ((chW % 8) != 0 && chN > 1)
    {
      expand_PIC (picbuf, expbuf, chN, chW, chH);
      shrink_PIC (expbuf, picbuf, chN, chW, chH);
    }
  xbm = xbmbuf;
  sprintf (xbm, "#define %s_width %d\n#define %s_height %d\nstatic char %s_bits[] = {\0", xbmname, chW * chN, xbmname, chH, xbmname);
  for (i = 0; i < ((chN * chW + 7) / 8) * chH; i++)
    {
      if (i % 16 == 0)
	sprintf (xbm + strlen (xbm), "\n\0");
      sprintf (xbm + strlen (xbm), "0x%02x,\0", (unsigned char) picbuf[i]);
    }
  sprintf (xbm + strlen (xbm) - 1, "};\n\0");
  return (int) strlen (xbm);
}

static int
PIC_to_IMG (picbuf, gifbuf, chN, chW, chH, color)
     char *picbuf, *gifbuf;
     int chN, chW, chH;
     int color;
{
#ifdef BLACKTXTASXBM
  if (!color)
    {
      isXBM = 1;
      return PIC_to_XBM (picbuf, gifbuf, chN, chW, chH, color);
    }
  else
    isXBM = 0;
#endif
  return PIC_to_GIF (picbuf, gifbuf, chN, chW, chH, color);
}

static int
getCII_H1 (picbuf, gifbuf, color, chN, chW, chH, isU)
     char *picbuf, *gifbuf;
     int color, chN, chW, chH, isU;
{
  char picbuf2[34 * (GIFWIDTH + 4)];
  int i, j, w;

  w = chN;
  for (i = 0; i < chH; i++)
    {
      expand_H1 (picbuf + i * w, picbuf2 + i * 2 * w * 2, w, chW);
      memcpy (picbuf2 + (i * 2 + 1) * w * 2, picbuf2 + i * 2 * w * 2, w * 2);
      if (i > 0)
	{
	  for (j = 0; j < w * 2; j++)
	    {
	      picbuf2[i * 2 * w * 2 + j] |= picbuf2[(i * 2 - 1) * w * 2 + j];
	    }
	}
    }

  chH *= 2;
  if (isU)
    {
      memset (picbuf2 + w * 2 * (chH + 0), 0x00, w * 2);
      memset (picbuf2 + w * 2 * (chH + 1), 0xff, w * 2);
      chH += 2;
    }

  return PIC_to_IMG (picbuf2, gifbuf, chN, chW * 2, chH, color);
}


static int
getCII (picbuf, gifbuf, color, chN, chW, chH, isU)
     char *picbuf, *gifbuf;
     int color, chN, chW, chH, isU;
{
  int w;
  if (isU)
    {
      w = chN;
      memset (picbuf + w * (chH + 0), 0x00, w);
      memset (picbuf + w * (chH + 1), 0xff, w);
      chH += 2;
    }
  return PIC_to_IMG (picbuf, gifbuf, chN, chW, chH, color);
}

static int
getCII_BOLD (picbuf, gifbuf, color, chN, chW, chH, isU)
     char *picbuf, *gifbuf;
     int color, chN, chW, chH, isU;
{
  int i, j, bit0, bit7, w;
  unsigned char *pb;

  pb = (unsigned char *) picbuf;
  w = chN;
  for (i = 0; i < chH; i++)
    {
      bit0 = 0;
      for (j = 0; j < w; j++)
	{
	  bit7 = *pb & 0x80;
	  *pb |= (*pb << 1) | bit0;
	  bit0 = bit7 >> 7;
	  pb++;
	}
    }
  if (isU)
    {
      memset (picbuf + w * (chH + 0), 0x00, w);
      memset (picbuf + w * (chH + 1), 0xff, w);
      chH += 2;
    }
  return PIC_to_IMG (picbuf, gifbuf, chN, chW, chH, color);
}

CII_getGIF (name, gifbuf)
     char *name, *gifbuf;
{
  int chW, chH;
  char charset[32], codebuf[(GIFWIDTH + 1) * 4 + 20], *code, *curgif[GIFWIDTH + 1],
    picbuf[(GIFWIDTH + 1) * 18], *tmpcurgif;
  static char uspcbuf[50];
  int color, isU, Hsize, chN, i, j, cc_id;

  code = codebuf;
#ifdef BLACKTXTASXBM
  xbmname = strrchr (name, '/') + 1;
#endif
  sscanf (name, "%[^/]/%6x%s", charset, &color, code);
  if ((cc_id = CharsetID (charset)) == -1)
    return 0;


  Hsize = *code++;
  if (Hsize < 'h' || Hsize > 'k')
    {
      return 0;
    }

  if (*code == 'u')
    {
      isU = 1;
      code++;
    }
  else
    {
      isU = 0;
    }

  chW = CII_CC[cc_id].width;
  chH = CII_CC[cc_id].height;



  for (chN = 0; *code; chN++, code += 4)
    {
      unsigned int c;

      if (chN == ((Hsize == 'h') ? GIFWH1 : GIFW))
	{
	  return 0;
	}
      if (!strncmp (code, "USPC", 4))
	{
	  curgif[chN] = uspcbuf;
	  continue;
	}
      sscanf (code, "%4x", &c);

      if (!CII_CC[cc_id].has8bit)
	{
	  c -= ((c % 256) / 128) ? 128 : 0;
	  c -= ((c / 256) / 128) ? 128 << 8 : 0;
	}
      if (c < CII_CC[cc_id].first || c > CII_CC[cc_id].last || CII_CC[cc_id].xbmind[c - CII_CC[cc_id].first] == -1)
	return 0;
      tmpcurgif = curgif[chN] = CII_CC[cc_id].xbmsrc[CII_CC[cc_id].xbmind[c - CII_CC[cc_id].first]];
      tmpcurgif++;
      if (chW > 8)
	{
	  curgif[++chN] = tmpcurgif;
	}
    }


  for (j = 0; j < chH; j++)
    {
      for (i = 0; i < chN; i++)
	{
	  picbuf[chN * j + i] = (chW > 8) ? curgif[i][j * 2] : curgif[i][j];
	}
    }
  if (chW > 8)
    chW /= 2;
  switch (Hsize)
    {
    case 'h':
      return getCII_H1 (picbuf, gifbuf, color, chN, chW, chH, isU);
    case 'i':
      return getCII_BOLD (picbuf, gifbuf, color, chN, chW, chH, isU);
    case 'j':
      return getCII_BOLD (picbuf, gifbuf, color, chN, chW, chH, isU);
    case 'k':
      return getCII (picbuf, gifbuf, color, chN, chW, chH, isU);
    }
}


CII_putHTTP_GIF (tc, vno, gkan)
     FILE *tc;
     char *gkan;
{
  int size;
  int leng;
  int chW;
  char gifbuf[(GIFWIDTH + 4) * (34 * 5 + 10) + (GIFWIDTH + 4) * 34 * 5 / 16 + 100];

  if ((size = CII_getGIF (gkan, gifbuf)) == 0)
    return 0;
  if (isXBM)
    leng = HTTP_putHeader (tc, vno, "image/x-xbitmap", size);
  else
    leng = HTTP_putHeader (tc, vno, "image/gif", size);
  leng += fwrite (gifbuf, 1, size, tc);
  return leng;
}
