/************************************************************************
* Copyright (c) 1995 Valentin Shopov, Blagoevgrad, Bulgaria             *
*                                                                       *
* Permission to use, copy, modify, and distribute this material for any *
* purpose and without fee is hereby granted, provided that the above    *
* copyright notice and this permission notice appear in all copies, and *
* that the name of the author not be used in advertising or publicity   *
* pertaining to this material without the specific, prior written       *
* permission of the author.                                             *
*                                                                       *
* THE AUTHOR MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY *
* OF THIS MATERIAL FOR ANY PURPOSE. IT IS PROVIDED "AS IS", WITHOUT ANY *
* EXPRESS OR IMPLIED WARRANTIES.                                        *
*************************************************************************/

/**************************ciilib1.5.1***********************************
 *                                                                      *
 *        Character to Inline Image Conversion Library (CIILIB)         *
 *                                                                      *
 *         	  x16 and x14 CII Fonts Builder.			*
 *                      by Valentin Shopov                              *
 *                      valentin@rs.aubg.bg                             *
 ************************************************************************/


#include <stdio.h>
#include "cii.h"

#define VERSION "0.2"

#define A_7bit  "7bit"
#define A_8bit  "8bit"
#define A_1byte "1byte"
#define A_2byte "2byte"

#define DELIMITER "--\r\n"

extern struct CharCodesXBM CII_CC[];
extern int CII_CC_SIZE;


/* imported from ciiserv.c */
static char *
Sprintf (out, fmt, a, b, c, d, e, f, g)
     char *out, *fmt, *a, *b, *c, *d, *e, *f, *g;
{
  sprintf (out, fmt, a, b, c, d, e, f, g);
  return out + strlen (out);
}

/* imported from ciiserv.c */
CII_putFont1_2 (csx, tc)
     struct CharCodesXBM *csx;
/*   char *charset; */
     FILE *tc;
{
  int leng;
/* struct CharCodesXBM *csx; */
  char head[2048], *hp;
  int map, mapx;
  char *bp;
  int nx, xi, cx, nb, bsize, bi;

/* csx = CharSetXBM (charset); */
  if (csx->name == 0)
    return 0;

  hp = head;
  hp = Sprintf (hp, "CII-Version: %s\n", VERSION);
  hp = Sprintf (hp, "Name: %s,%s\n", csx->regname, csx->name);
  hp = Sprintf (hp, "Size: %d,%d\n", csx->width, csx->height);
  hp = Sprintf (hp, "Range: %d,%d\n", csx->first, csx->last);
  hp = Sprintf (hp, "Attrs: ");
  if (csx->is2bytes)
    hp = Sprintf (hp, A_2byte);
  else
    hp = Sprintf (hp, A_1byte);
  hp = Sprintf (hp, ",");
  if (csx->has8bit)
    hp = Sprintf (hp, A_8bit);
  else
    hp = Sprintf (hp, A_7bit);
  hp = Sprintf (hp, "\n");
  hp = Sprintf (hp, "\f\n\n");
  fputs (head, tc);
  leng = strlen (head);

  nx = csx->last - csx->first + 1;
  map = 0;
  mapx = 0;
  nb = 0;
  for (xi = 0; xi < nx; xi++)
    {
      if (0 <= csx->xbmind[xi])
	{
	  nb++;
	  map |= (1 << mapx);
	}
      if (++mapx == 8)
	{
	  putc (map, tc);
	  leng++;
	  map = 0;
	  mapx = 0;
	}
    }
  if (mapx != 0)
    {
      putc (map, tc);
      leng++;
    }
  bsize = (csx->width + 7) / 8 * csx->height;
  nb = 0;
  for (xi = 0; xi < nx; xi++)
    {
      if ((cx = csx->xbmind[xi]) < 0)
	continue;
      bp = csx->xbmsrc[cx];
      for (bi = 0; bi < bsize; bi++)
	putc (bp[bi], tc);
      nb++;
    }
  leng += nb * bsize;
  return leng;
}


main (ac, av)
     char *av[];
{
  int i, leng, leng1;
  FILE *ciif;
  struct CharCodesXBM *csx;


  if (ac < 2)
    {
      fprintf (stderr, "Usage: %s  <CII Fonts File Full Path>\n", av[0]);
      exit (1);
    }


  ciif = fopen (av[1], "w+");
  if (ciif == NULL)
    {
      fprintf (stderr, "Error opening \"%s\"!\n", av[1]);
      exit (1);
    }

  for (i = 0, leng = 0; CII_CC[i].regname; i++)
    {
      csx = &CII_CC[i];
      leng += leng1 = CII_putFont1_2 (csx, ciif);
      fputs (DELIMITER, ciif);
      fprintf (stderr, "%s: %d (%s)\n", csx->regname, leng1, csx->name);

    }
  fclose (ciif);
  fprintf (stderr, "total: %d\n", leng);
  exit (0);

}
